// Fixed filter bank neural network.
// Sean O'Connor
class FFBNet {
  int vecLen;
  int depth;
  float scale;
  float rate;
  float[] params;
  byte[] signs;
  float[] work;
  float[] flips;

  FFBNet(int vecLen, int depth, float rate) {
    this.vecLen = vecLen;
    this.depth = depth;
    scale = 1.0 / sqrt(vecLen);
    this.rate = rate * this.scale;
    params = new float[2 * vecLen * depth];
    signs = new byte[vecLen * depth];
    work = new float[vecLen];
    flips = new float[vecLen];
    for (int i = 0; i < vecLen; i++) {
      flips[i] = sin(i * 1.895) < 0 ? -scale : scale;
    }
    for (int i = 0; i < params.length; i++) {
      this.params[i] = scale;
    }
  }

  void recall(float[] result, float[] input) {
    for (int i = 0; i < vecLen; i++) {
      result[i] = input[i] * flips[i];
    }
    int signIdx = 0; // parameter index
    for (int i = 0; i < depth; i++) {
      wht(result);
      for (int j = 0; j < vecLen; j++) {
        int signBit = result[j] < 0 ? 0 : 1;
        this.signs[signIdx] = (byte)signBit;
        result[j] *= params[2 * signIdx + signBit];
        signIdx++;
      }
    }
    wht(result);
  }

  void train(float[] target, float[] input) {
    recall(work, input);
    for (int i = 0; i < this.vecLen; i++) {
      work[i] = (target[i] - work[i]) * rate;
    }
    int signsIdx = this.vecLen * this.depth - 1;
    for (int i = 0; i < this.depth; i++) {
      wht(this.work);
      for (int j = this.vecLen - 1; j >= 0; j--) {
        int b = this.signs[signsIdx];
        float v = 2 * b - 1;
        params[2 * signsIdx + b] += v * work[j];
        work[j] *=params[2 * signsIdx + b] < 0f ? -scale : scale;
        signsIdx--;
      }
    }
  }
}

// Fast Walsh Hadamard Transform
void wht(float[] vec) {
  final int n = vec.length;
  int hs = 1;
  while (hs < n) {
    int i = 0;
    while (i < n) {
      final int j = i + hs;
      while (i < j) {
        float a = vec[i];
        float b = vec[i + hs];
        vec[i] = a + b;
        vec[i + hs] = a - b;
        i += 1;
      }
      i += hs;
    }
    hs += hs;
  }
}

// Sum of squared difference cost
float costL2(float[] vec, float[] tar) {
  float cost = 0;
  for (int i = 0; i < vec.length; i++) {
    float e = vec[i] - tar[i];
    cost += e * e;
  }
  return cost;
}

// Test with Lissajous curves
color c1;
color c2;
float[][] ex = new float[8][256];
float[] work = new float[256];
FFBNet parentNet;

void setup() {
  size(400, 400);
  parentNet = new FFBNet(256, 6, 0.00002);
  c1 = #FFCC00;
  c2 = #FFCCCC;
  for (int i = 0; i < 127; i++) {
    // Training data
    float t = (i * 2 * PI) / 127;
    ex[0][2 * i] = sin(t);
    ex[0][2 * i + 1] = sin(2 * t);
    ex[1][2 * i] = sin(2 * t);
    ex[1][2 * i + 1] = sin(t);
    ex[2][2 * i] = sin(2 * t);
    ex[2][2 * i + 1] = sin(3 * t);
    ex[3][2 * i] = sin(3 * t);
    ex[3][2 * i + 1] = sin(2 * t);
    ex[4][2 * i] = sin(3 * t);
    ex[4][2 * i + 1] = sin(4 * t);
    ex[5][2 * i] = sin(4 * t);
    ex[5][2 * i + 1] = sin(3 * t);
    ex[6][2 * i] = sin(2 * t);
    ex[6][2 * i + 1] = sin(5 * t);
    ex[7][2 * i] = sin(5 * t);
    ex[7][2 * i + 1] = sin(2 * t);
  }
  textSize(16);
}

void draw() {
  background(0);
  for (int i = 0; i < 100; i++) {  
    for (int j = 0; j < 8; j++) {
      parentNet.train(ex[j], ex[(j + 1) % ex.length]);
    }
  }
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 255; j += 2) {
      set(25 + i * 40 + int(18 * ex[i][j]), 44 + int(18 * ex[i][j + 1]), c1);
    }
  }
  for (int i = 0; i < 8; i++) {
    parentNet.recall(work, ex[i]);
    for (int j = 0; j < 255; j += 2) {
      set(25 + i * 40 + int(18 * work[j]), 104 + int(18 * work[j + 1]), c2);
    }
  }
  text("Training Data", 5, 20);
  text("Recall", 5, 80);
  text("Iterations: " + frameCount, 5, 150);
}
