
class FFBNet {
  int vecLen;
  int depth;
  float scale;
  float[] params;
  float[] flips;
  
  FFBNet(int vecLen,int depth) {
    this.vecLen = vecLen;
    this.depth = depth;
    scale = 1f / sqrt(vecLen);
    params = new float[2 * vecLen * depth];
    flips = new float[vecLen];
    for (int i = 0; i < vecLen; i++) {
      flips[i] = sin(i * 1.895) < 0 ? -scale : scale;
    }
    for (int i = 0; i < params.length; i++) {
      params[i] = scale;
    }
  }

  void recall(float[] result, float[] input) {
    for (int i = 0; i < vecLen; i++) {
      result[i] = input[i] * flips[i];
    }
    int paramIdx = 0;
    for (int i = 0; i < depth; i++) {
      wht(result);
      for (int j = 0; j < vecLen; j++) {
        int signBit = result[j] < 0 ? 0 : 1;
        result[j] *= params[paramIdx + signBit];
        paramIdx += 2;
      }
    }
    wht(result);
  }
}

// Fast Walsh Hadamard Transform
void wht(float[] vec) {
  final int n = vec.length;
  int hs = 1;
  while (hs < n) {
    int i = 0;
    while (i < n) {
      final int j = i + hs;
      while (i < j) {
        float a = vec[i];
        float b = vec[i + hs];
        vec[i] = a + b;
        vec[i + hs] = a - b;
        i += 1;
      }
      i += hs;
    }
    hs += hs;
  }
}

// Sum of squared difference cost
float costL2(float[] vec, float[] tar) {
  float cost = 0;
  for (int i = 0; i < vec.length; i++) {
    float e = vec[i] - tar[i];
    cost += e * e;
  }
  return cost;
}

class Mutator {
  float[] previous;
  int[] pIdx;
  float precision;
  float limit;
  Mutator(int size, float precision, float limit) {
    previous = new float[size];
    pIdx = new int[size];
    this.precision = precision;
    this.limit = limit;
  }
  void mutate(float[] vec) {
    for (int i = 0; i < previous.length; i++) {
      int rpos = int(random(vec.length));
      float v = vec[rpos];
      pIdx[i] = rpos;
      previous[i] = v;
      float m = 2 * limit * exp(random(-precision, 0));
      if (random(1) < 0.5) m = -m;
      float vm = v + m;
      if (vm >= limit) vm = v;
      if (vm <= -limit) vm = v;
      vec[rpos] = vm;
    }
  }
  void undo(float[] vec) {
    for (int i = previous.length - 1; i >= 0; i--) {
      vec[pIdx[i]] = previous[i];
    }
  }
}

// Test with Lissajous curves
color c1;
color c2;
float[][] ex = new float[8][256];
float[] work = new float[256];
float parentCost = Float.POSITIVE_INFINITY;
FFBNet parentNet;
Mutator mut;

void setup() {
  size(400, 400);
  parentNet = new FFBNet(256, 6);
  mut = new Mutator(10, 25, 2 * parentNet.scale);
  c1 = #FFCC00;
  c2 = #FFCCCC;
  for (int i = 0; i < 127; i++) {
    // Training data
    float t = (i * 2 * PI) / 127;
    ex[0][2 * i] = sin(t);
    ex[0][2 * i + 1] = sin(2 * t);
    ex[1][2 * i] = sin(2 * t);
    ex[1][2 * i + 1] = sin(t);
    ex[2][2 * i] = sin(2 * t);
    ex[2][2 * i + 1] = sin(3 * t);
    ex[3][2 * i] = sin(3 * t);
    ex[3][2 * i + 1] = sin(2 * t);
    ex[4][2 * i] = sin(3 * t);
    ex[4][2 * i + 1] = sin(4 * t);
    ex[5][2 * i] = sin(4 * t);
    ex[5][2 * i + 1] = sin(3 * t);
    ex[6][2 * i] = sin(2 * t);
    ex[6][2 * i + 1] = sin(5 * t);
    ex[7][2 * i] = sin(5 * t);
    ex[7][2 * i + 1] = sin(2 * t);
  }
  textSize(16);
}

void draw() {
  background(0);
  for (int i = 0; i < 100; i++) {
    mut.mutate(parentNet.params);
    float cost = 0;
    for (int j = 0; j < 8; j++) {
      parentNet.recall(work, ex[(j + 1) % ex.length]);
      cost += costL2(work, ex[j]);
    }
    if (cost < parentCost) {
      parentCost = cost;
    } else {
      mut.undo(parentNet.params);
    }
  }
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 255; j += 2) {
      set(25 + i * 40 + int(18 * ex[i][j]), 44 + int(18 * ex[i][j + 1]),c1);
    }
  }
  for (int i = 0; i < 8; i++) {
    parentNet.recall(work, ex[i]);
    for (int j = 0; j < 255; j += 2) {
      set(25 + i * 40 + int(18 * work[j]), 104 + int(18 * work[j + 1]),c2);
    }
  }
  text("Training Data", 5, 20);
  text("Recall", 5, 80);
  text("Iterations: " + frameCount, 5, 150);
  text("Cost: " + parentCost, 5, 170);
}
