// Fixed filter bank neural network.
// Sean O'Connor
class FFBNet {
  // vecLen must be 2,4,8,16,32.....
  constructor(vecLen, depth) {
    this.vecLen = vecLen;
    this.depth = depth;
    this.scale = 1.0 / sqrt(vecLen);
    this.params = new Float32Array(2 * vecLen * depth);
    this.flips = new Float32Array(vecLen);
    for (let i = 0; i < vecLen; i++) {
      this.flips[i] = sin(i * 1.895) < 0 ? -this.scale : this.scale;
    }
    for (let i = 0; i < this.params.length; i++) {
      this.params[i] = this.scale;
    }
  }

  recall(result, input) {
    for (let i = 0; i < this.vecLen; i++) {
      result[i] = input[i] * this.flips[i];
    }
    let paramIdx = 0; // parameter index
    for (let i = 0; i < this.depth; i++) {
      wht(result);
      for (let j = 0; j < this.vecLen; j++) {
        const signBit = result[j] < 0 ? 0 : 1;
        result[j] *= this.params[paramIdx + signBit];
        paramIdx += 2;
      }
    }
    wht(result);
  }
}

// Fast Walsh Hadamard Transform
function wht(vec) {
  const n = vec.length;
  let hs = 1;
  while (hs < n) {
    let i = 0;
    while (i < n) {
      const j = i + hs;
      while (i < j) {
        var a = vec[i];
        var b = vec[i + hs];
        vec[i] = a + b;
        vec[i + hs] = a - b;
        i += 1;
      }
      i += hs;
    }
    hs += hs;
  }
}

class Mutator {
  constructor(size, precision, limit) {
    this.previous = new Float32Array(size);
    this.pIdx = new Int32Array(size);
    this.precision = precision;
    this.limit = limit;
  }
  mutate(vec) {
    for (let i = 0; i < this.previous.length; i++) {
      let rpos = int(random(vec.length));
      let v = vec[rpos];
      this.pIdx[i] = rpos;
      this.previous[i] = v;
      let m = 2 * this.limit * exp(random(-this.precision, 0));
      if (random() < 0.5) m = -m;
      let vm = v + m;
      if (vm >= this.limit) vm = v;
      if (vm <= -this.limit) vm = v;
      vec[rpos] = vm;
    }
  }
  undo(vec) {
    for (let i = this.previous.length - 1; i >= 0; i--) {
      vec[this.pIdx[i]] = this.previous[i];
    }
  }
}

function makeMaze() {
  let m = new Int8Array(mazeEdge * mazeEdge);
  let xy = random() < 0.5 ? true : false;
  let x = mazeMid;
  let y = mazeMid;
  for (let j = 0; j < mazeLines; j++) {
    let r = int(random(1, mazeEdge - 1));
    if (xy) {
      let sx = min(r, x);
      let ex = max(r, x);
      for (; sx <= ex; sx++) m[sx + y * mazeEdge] = 1;
      x = r;
    } else {
      let sy = min(r, y);
      let ey = max(r, y);
      for (; sy <= ey; sy++) m[x + sy * mazeEdge] = 1;
      y = r;
    }
    xy = !xy;
  }
  let r = random() < 0.5 ? 0 : 48;
  if (xy) {
    let sx = min(r, x);
    let ex = max(r, x);
    for (; sx <= ex; sx++) m[sx + y * mazeEdge] = 1;
  } else {
    let sy = min(r, y);
    let ey = max(r, y);
    for (; sy <= ey; sy++) m[x + sy * mazeEdge] = 1;
  }
  return m;
}

function move(maze, x, y, neuralNet, vec) {
  dir[0] = 0;
  dir[1] = 0;
  if (x === 0 || x === mazeEdge - 1 || y === 0 || y === mazeEdge - 1) return; // Game finished with a win.
  let mp = x + y * mazeEdge;
  vec[0] = maze[mp + 1];
  vec[1] = maze[mp - 1];
  vec[2] = maze[mp + mazeEdge];
  vec[3] = maze[mp - mazeEdge];
  neuralNet.recall(vec, vec);
  let max = vec[0];
  let pos = 0;
  for (let i = 1; i < 4; i++) {
    if (vec[i] > max) {
      max = vec[i];
      pos = i;
    }
  }
  switch (pos) {
    case 0:
      if (maze[mp + 1] === 1) dir[0] = 1;
      break;
    case 1:
      if (maze[mp - 1] === 1) dir[0] = -1;
      break;
    case 2:
      if (maze[mp + mazeEdge] === 1) dir[1] = 1;
      break;
    case 3:
      if (maze[mp - mazeEdge] === 1) dir[1] = -1;
  }
}

let neuralNet;
let parentCost = Number.POSITIVE_INFINITY;
let mut;
let work = new Float32Array(32);
let dir = [0, 0];
let mazeMid; // Mid-point of maze
const mazeEdge = 49; // Maze edge length must be odd.
const mazeLines = 25; // Complexity of maze
const gameLimit = 300;
const nMazes = 200;

let gameMaze; // new maze not part of training set.
let gameX; // Current x,y
let gameY;
let gameCounter = 0; //Current move number in game
let mazes = [];
let train = false;

function setup() {
  createCanvas(490, 490);
  neuralNet = new FFBNet(32, 8);
  mut = new Mutator(10, 30, 2 * neuralNet.scale);
  mazeMid = int(mazeEdge / 2);
  for (let i = 0; i < nMazes; i++) {
    mazes[i] = makeMaze();
  }
}

function draw() {
  if (train) {
    mut.mutate(neuralNet.params);
    let childCost = 0;
    for (let i = 0; i < nMazes; i++) {
      let x = mazeMid;
      let y = mazeMid;
      for (let j = 0; j < work.length; j++) {
        work[j] = 0;
      }
      for (let j = 0; j < gameLimit; j++) {
        move(mazes[i], x, y, neuralNet, work);
        x += dir[0];
        y += dir[1];
        if (x === 0 || x === mazeEdge - 1 || y === 0 || y === mazeEdge - 1) {
          childCost -= 1000000; // Found way out
          break;
        }
        childCost -= dist(mazeMid, mazeMid, x, y);
      }
    }
    if (childCost < parentCost) {
      parentCost = childCost;
    } else {
      mut.undo(neuralNet.params);
    }
    background("yellow");
    fill(0);
    text("Cost: " + parentCost.toFixed(0), 10, 30);
  } else {
    background(0);
    if (gameCounter === gameLimit) {
      gameCounter = 0;
    }
    if (gameCounter === 0) {
      gameMaze = makeMaze();
      gameX = mazeMid;
      gameY = mazeMid;
      for (let i = 0; i < work.length; i++) {
        work[i] = 0;
      }
    }
    gameCounter += 1;
    move(gameMaze, gameX, gameY, neuralNet, work);
    gameX += dir[0];
    gameY += dir[1];

    fill(color("green"));
    noStroke();
    for (let i = 0; i < mazeEdge; i++) {
      for (let j = 0; j < mazeEdge; j++) {
        if (gameMaze[j + i * mazeEdge] === 0) {
          square(j * 10, i * 10, 10);
        }
      }
    }
    fill(color("yellow"));
    circle(5 + gameX * 10, 5 + gameY * 10, 8);
  }
}

function mouseClicked() {
  train = !train;
  gameCounter = 0;
}
